# -*- coding: utf-8 -*-

# SAX
from xml.sax import ContentHandler
from xml.sax import make_parser
from xml.sax import parseString

# para el cliente http
import httplib

from pandac.PandaModules import Point3

from direct.task.Task import Task

#from pandac.PandaModules import HTTPClient
#from pandac.PandaModules import Ramfile
#from pandac.PandaModules import URLSpec

from escalas import Escalas
from Dijkstra import shortest_path

# define los limites del mapa
class Bounds:
    
    """
        Coordenadas geográficas
    """
    def setMinCoord(self,coord):
        self.mincoord = coord
    def setMaxCoord(self,coord):
        self.maxcoord = coord
    
    def getMinCoord(self):
        return self.mincoord
    def getMaxCoord(self):
        return self.maxcoord
    
    """
        Coordenadas cartesianas
    """
    def setMinPos(self,p):
        self.minpos = p
    def getMinPos(self):
        return self.minpos

    def setMaxPos(self,p):
        self.maxpos = p
    def getMaxPos(self):
        return self.maxpos

# un Nodo es una posicion concreta del mapa
class Nodo:
    
    def __init__(self):
        
        # diccionario de etiquetas del nodo
        self.Tags = {}
        # lista de way donde el nodo aparece
        self.Ways = []
        
        # indica si esta en una carretera de un solo carril
        self.oneway = 'yes'
    
        # nombre de way al que pertenece el nodo
        self.wayName = []
        
    def setDatos(self,id,lon,lat):
        self.n = (id,lon,lat)     
        
    def getId(self):
        return self.n[0]
    def getLon(self):
        return self.n[1]
    def getLat(self):
        return self.n[2]
    
    # incluye una nueva tupla (key,value) a la lista de tags
    def addTag(self,k,v):
        self.Tags[k] = v
    def addTags(self,t):
        self.Tags = t
        # lista de tags de way
    def getTags(self):
        return self.Tags
    
    # si way tiene la clave la devuelve, None en caso contrario
    def getTagValue(self,tag):
        if self.Tags.has_key(tag):
            return self.Tags[tag]
        else:
            #print "no existe la eiqueta " + str(tag)
            return None
    # comprueba si way tiene la clave    
    def has_tag(self,tag):
        return self.Tags.has_key(tag)

    def setPos(self,p):
        self.pos = p
    def getPos(self):
        return self.pos
    def getX(self):
        return self.pos.getX()
    def getY(self):
        return self.pos.getY()
    
    def setOneWay(self,o):
        self.oneway = o
    def getOneWay(self):
        return self.oneway
    
    def setWayName(self,wn):
        if not self.wayName.__contains__(wn):
            self.wayName.append(wn)
        
    def getWayName(self):
        if self.wayName.__len__()==0:
            return ''
        else:
            
            str = self.wayName[0]
            for s in self.wayName[1:]:
                str = str + ' con ' + s
            
            return str
    
    """
    Se utiliza en la clase escenario para saber que way pertenecen
    a los nodos dentro la porcion de mapa a pintar
    """
    def addWay(self,w):
        self.Ways.append(w)
    def getListaWays(self):
        return self.Ways

# Describe cualquier elemento que pueda aparecer en el mapa
class Way:
    
    def __init__(self):
        
        # id de way
        self.id = None
        # lista de ref de way
        self.Refs = []
        # diccionario de tags, {Key:Value}
        self.Tags = {}
        
        # identifica si es una carretera, un parque, una plaza ...
        self.tipo = None

    # establece el valor id
    def setId(self,i):
        self.id = i
    # incluye una nueva referencia a la lista de referencias
    def addRef(self,r):
        self.Refs.append(r)
    # incluye una nueva tupla (key,value) a la lista de tags
    def addTag(self,k,v):
        self.Tags[k] = v 
    def addTags(self,t):
        self.Tags = t 
    # devuelve id de way
    def getId(self):
        return self.id
    # lista de referencias de way
    def getListaRefs(self):
        return self.Refs
    # lista de tags de way
    def getTags(self):
        return self.Tags
    
    # si way tiene la clave la devuelve, None en caso contrario
    def getTagValue(self,tag):
        if self.Tags.has_key(tag):
            return self.Tags[tag]
        else:
            #print "no existe la eiqueta " + str(tag)
            return None
    # comprueba si way tiene la clave    
    def has_tag(self,tag):
        return self.Tags.has_key(tag)

    def setTipo(self,x):
        self.tipo = x
    def getTipo(self):
        return self.tipo

# Describe la matriz de camino del mapa
class MatrizCamino():
    
    def __init__(self,nodos,carreteras):
        
        print "Matriz de camino"
        
        self.MatrizCamino = {}
        
        for w in carreteras:
            
            # lista de nodos que contien way
            refs = w.getListaRefs()
            nrefs = refs.__len__()
            
            # determina si way es de doble sentido
            if w.getTagValue('oneway') == 'yes' or w.getTagValue('oneway') == '1':
                self.oneway = 'yes'
            else:
                self.oneway = 'no'
            
            self.wayName = w.getTagValue('name')
            if self.wayName == None:
                if w.getTagValue('ref') != None:
                    self.wayName = w.getTagValue('ref')
                    
            i=0
            while i < (nrefs-1) :
                 
                if not self.MatrizCamino.has_key( refs[i] ):
                    self.MatrizCamino[ refs[i] ] = []
                    
                # incluye en la lista que estan delante al siguiente nodo
                self.MatrizCamino[ refs[i] ].append( refs[i+1] )
                
                # si way es de doble sentido crea el enlace inverso
                if self.oneway == 'no':
                    
                    # si el nodo no esta en el diccionario lo registra
                    if not self.MatrizCamino.has_key( refs[i+1] ):
                        self.MatrizCamino[ refs[i+1] ] = []
                    
                    self.MatrizCamino[ refs[i+1] ].append( refs[i] )
                
                    
                if self.wayName != None:
                    nodos[ refs[i] ].setWayName(self.wayName)
                if self.oneway == 'no':
                    nodos[ refs[i] ].setOneWay(self.oneway)
            
                
                i = i+1
            # end-while
                
            if self.oneway == 'yes':
                if not self.MatrizCamino.has_key( refs[i] ):
                    self.MatrizCamino[ refs[i] ] = []
                
            if self.wayName != None:
                nodos[ refs[i] ].setWayName(self.wayName)
            if self.oneway == 'no':
                nodos[ refs[i] ].setOneWay(self.oneway)
    
    def getMatriz(self):
        return self.MatrizCamino
        
    def getListaNodos(self):
        return self.MatrizCamino.keys()
            
    def getListaAdyacentes(self,id):
        return self.MatrizCamino[id]
    
    def has_nodo(self,nd):
        return self.MatrizCamino.has_key(nd)
        
# clase que gestiona los eventos produccidos por sax
class Manejador(ContentHandler):
    
    def __init__(self,bounds,nodos,ways,carreteras):
        
        # copia local de bounds
        self.B = bounds
        # copia local del diccionario de nodos
        self.N = nodos
        # copia local de ways, objeto que contiene la lista de ways
        self.W = ways
        # nodo auxiliar para almacenar los datos que contiene
        self.nodo_aux = Nodo()
        #way auxuliar para almacenar los datos de way mientras se termina de leer
        self.way_aux = Way()
        # para almacenar temporalmente los tags
        self.tags_aux = {}
    
        # diccionario que agrupo los elementos de una misa clase en listas
        self.listaCarreteras = carreteras
    
        # contiene los valores de la clave highway que se corresponden con una carretera
        self.tagsCarreteras = ['motorway','motorway_link','trunk','trunk_link','primary','primary_link',
                           'secondary','secondary_link','tertiary','unclassified','road','residential',
                           'living_street','track','roundabout']
        
    # se ejecuta la comiezo de cada elemento
    def startElement(self,nombre,atributos):
    
        if nombre == 'bounds':

            self.B.setMinCoord(Point3(float(atributos.get('minlon')),
                                      float(atributos.get('minlat')),
                                      0))
            
            self.B.setMinPos(Point3(0,0,0))
            
            self.B.setMaxCoord(Point3(float(atributos.get('maxlon')),
                               float(atributos.get('maxlat')),
                               0))
            
            self.B.setMaxPos( self.B.getMaxCoord()*Escalas['gr-px'] )
        
        if nombre == 'node':
            self.nodo_aux.setDatos(atributos.get('id'), 
                                   float(atributos.get('lon')),
                                   float(atributos.get('lat')))
        
            self.nodo_aux.setPos((Point3(self.nodo_aux.getLon(),
                                        self.nodo_aux.getLat(),
                                        0.0) - self.B.getMinCoord()) * Escalas['gr-px'])            
            
        if nombre == 'way':
            # crea un objeto way auxiliar con su id
            self.way_aux.setId( atributos.get('id') )
            
        if nombre == 'nd':
            # incluye una nueva referencia al way auxiliar
            self.way_aux.addRef(atributos.get('ref'))
            self.N[atributos.get('ref')].addWay(self.way_aux.getId())
        
        if nombre == 'tag':
            # encode() transforma a utf-8 y evita problemas con los caracteres
#            self.Tags.append( (atributos.get('k'),atributos.get('v').encode('utf-8')) )
            self.tags_aux[atributos.get('k')] = atributos.get('v').encode('utf-8')                                  
                                 
    def endElement(self,nombre):
        
        if nombre == 'way':
            
            self.way_aux.addTags(self.tags_aux)
            self.tags_aux = {}
            
            # incluye el way con todos los datos(refs y tags) a la lista
            self.W[self.way_aux.getId()] = self.way_aux
            
            # comprueba si es una carretera
            if self.tagsCarreteras.__contains__(self.way_aux.getTagValue('highway')):
                self.way_aux.setTipo('carreteras')
                self.listaCarreteras.append(self.way_aux)
                
                
#            if self.way_aux.getTagValue('highway') == 'pedestrian':
#                if self.way_aux.getTagValue('area') == 'yes':
#                    self.dicElementos['plazas'].append(self.way_aux.getId())
            
            # crea un nuevo way
            self.way_aux = Way()
            
        if nombre == 'node':
            
            self.nodo_aux.addTags(self.tags_aux)
            self.tags_aux = {}
            
            self.N[ self.nodo_aux.getId() ] = self.nodo_aux
            
            self.nodo_aux = Nodo()
            
            
#class ClienteHTTP(threading.Thread):
#
#    def __init__(self,latitud,longitud):
#
#        threading.Thread.__init__(self)
#
#        conn = httplib.HTTPConnection("www.openstreetmap.org")
#        conn.request("GET",
#                     "/api/0.6/map?bbox=" + str(longitud-0.02) + "," + str(latitud-0.02) + "," + str(longitud+0.02) + "," + str(latitud+0.02))
#        r1 = conn.getresponse()
#        print r1.status, r1.reason # 200, OK
#        data1 = r1.read()
#        print data1
#        http = HTTPClient()
#
#        self.channel = http.getDocument(URLSpec("http://www.openstreetmap.org/api/0.6/map?bbox=" + str(longitud-0.02)
#                                           +"," 
#                                           + str(latitud-0.02)
#                                           +","
#                                           + str(longitud+0.02)
#                                           +","
#                                           +str(latitud+0.02)))
#        
#        
#        self.rf = Ramfile()
#        self.channel.downloadToRam(self.rf)
#        taskMgr.add(self.getHTTP, 'download')
#
#    def getHTTP(self,task):
#
#        if self.channel.run():
#            # Still waiting for file to finish downloading.
#            return task.cont
#        if not self.channel.isDownloadComplete():
#            print "Error downloading file."
#            return task.done
#        self.data1 = self.rf.getData()
#        print self.data1
##        parseString(self.data1,Manejador(self.bounds,self.nodos,self.ways))
#        return task.done



class Mapa():    

    
    def __init__(self,osm="http",escala=Escalas['gr-px'],latitud=36.7,longitud=-6.1):

        print "Mapa"

        # almacena el nombre del mapa
        self.mapa = osm
        # alamcena los limites del sistema
        self.bounds = Bounds()
        # diccionario de nodos por id
        self.nodos = {}  
        # Lista de ways
        self.ways = {}

        # lista de carreteras para la matriz de camino
        self.carreteras = []
    
        if self.mapa == "http":
    
            print 'Accediendo a www.openstreetmap.org ...'
            
            # mapa de http client
            conn = httplib.HTTPConnection("www.openstreetmap.org")
            conn.request("GET",
                         "/api/0.6/map?bbox=" + str(longitud-0.02) + "," + str(latitud-0.02) + "," + str(longitud+0.02) + "," + str(latitud+0.02))
            r1 = conn.getresponse()
            print r1.status, r1.reason # 200, OK
            data1 = r1.read()
            parseString(data1,Manejador(self.bounds,self.nodos,self.ways,self.carreteras))
              

        else:
            
            # crea un objeto manejador de eventos 
            manejador = Manejador(self.bounds,self.nodos,self.ways,self.carreteras)
            # crea un objeto parser
            parser = make_parser()
            # asigna nuestro manejador de evenos al parser
            parser.setContentHandler(manejador)
            # abre el fichero xml con los datos
            fichero = open('./mapas/' + self.mapa, 'r')
            # asigna al parser el fichero con la informacion
            parser.parse(fichero)
        
        print "Mapa:Documento osm procesado"
        
        # crea la martiz de camino
        self.Mcamino = MatrizCamino(self.nodos,self.carreteras)

    # devuelve el nombre del mapa
    def getMapa(self):
        return str(self.nombre)
    
    def CalcularPorcionMapa(self,pos,distancia):
        
        print 'Mapa:CalculaPorcionMapa'
        
        """
         Dtermina que elmentos estan en la parte
         visible de la porcion de mapa elegida
        """
        
        # contiene la listas  way clasificados
        self.porcion_mapa = {}
    
        # lista para los way que son carretera
        self.porcion_mapa['carreteras'] = []
    
        for w in self.ways.itervalues():
            
            if w.getTipo() == 'carreteras':
                
                for r in w.getListaRefs():
                    if (self.nodos[r].getPos() - pos).length() < distancia:
                        """
                         añade el objeto way a la lista, internamente son referencias
                         a objetos
                        """
                        self.porcion_mapa['carreteras'].append(w)
                        # no es necesario seguir recorriento la lista de referencias
                        break
           
           # OTROS TIPOS DE ELEMENTOS
#           if w.getTipo() == 'plazas':
#               pass
                
        """
         Determina la la parte de la matriz de camino
         perteneciente a la parte visible del mapa
        """
        
        self.porcion_camino = []
        
        for n in self.Mcamino.getMatriz().iterkeys():
            if (self.nodos[n].getPos() - pos).length() < distancia:
                self.porcion_camino.append(n)
        
        
    """
    Devuevle un diccionario de elementos  que estan dentro 
    de la pocion de mapa seleccionada
    """
    def getPorcionMapa(self):
      
        # devuelve la lista de elementos que deben ser pintados
        return self.porcion_mapa
    
    """
    devuelve una lisa de nodos pertenecientes a la matriz de camino
    que estan dentro de la porcion de mapa dibujada
    """
    def getPorcionCamino(self):
        return self.porcion_camino
    
    # devuelve un objeto bounds con los limites del mapa
    def getBounds(self):
        return self.bounds
    def getMinLon(self):
        return self.bounds.getMinLon()
    def getMaxLon(self):
        return self.bounds.getMaxLon()
    def getMinLat(self):
        return self.bounds.getMinLat()
    def getMaxLat(self):
        return self.bounds.getMaxLat()
    # devuelve una tupla con la posicion minima del limite del mapa
    def getBoundsMinX(self):
        return self.bounds.getMinX()
    def getBoundsMinY(self):
        return self.bounds.getMinY()
    # devuelve una tupla con la posicion maxima del limite del mapa
    def getBoundsMaxX(self):
        return self.bounds.getMaxX()
    def getBoundsMaxY(self):
        return self.bounds.getMaxY()    
        
    # devuelve la lista de id de los nodos del mapa
    def getListaNodosId(self):
        return self.nodos.iterkeys()
    # tupla con la posicon del nodo
    def getNodoPosById(self,id):
        return self.nodos[id].getPos()
    # devuelve la coordenada X de la posicion del nodo
    def getNodoXById(self,id):
        return self.nodos[id].getX()
    # devulve la coordenada Y de la posicion del nodo
    def getNodoYById(self,id):
        return self.nodos[id].getY()
    
    def getNodoLonById(self,id):
        return self.nodos[id].getLon() # longitud del nodo    
    def getNodoLatById(self,id):
        return self.nodos[id].getLat() # latitud del nodo
    
    # devuelve la lista de ways
    def getListaWays(self):
        return self.ways.keys()
    
    def getListaWaysByNodo(self,nd):
        return self.nodos[nd].getListaWays()
    
    def getWay(self,id):
        return self.ways[id]
    
    def getWayNombreByNodo(self,n):
        return self.nodos[n].getWayName()
    
    def getWayNombreById(self,w):
        return self.ways[w].getTagValue('name')
    
    def getOneWayByNodo(self,id):
        return self.nodos[id].getOneWay()    
    
    # metodos de la matriz de camino
    def getMatrizCamino(self):
        return self.Mcamino
    def getListaNodosMatrizCamino(self):
        return self.Mcamino.getListaNodos()
    
#    def getListaNodosSiguientes(self,id):
#        return self.Mcamino.getListaNodosSiguientes(id)
#    def getListaNodosAnteriores(self,id):
#        return self.Mcamino.getListaNodosAnteriores(id)
    
    def getListaAdyacentes(self,id):
        return self.Mcamino.getListaAdyacentes(id)

    def getRuta(self,origen,destino):
        
        print 'Mapa:getRuta'
        
        # matriz de costes
        self.G = {}
        
        # establece los costes iniciales
        for k in self.Mcamino.getListaNodos():
            aux  = {}
            for i in self.Mcamino.getListaAdyacentes(k):
                aux[i] = (self.getNodoPosById(k) - self.getNodoPosById(i)).length()
            self.G[k] = aux
        
        try:
            return shortest_path(self.G,origen,destino)
        except:
            print 'Mapa:getRuta, no fue posible encontrar un camino.'
            return []
        
    def getListaCarreteras(self):
        return self.carreteras

#from pandac.PandaModules import HTTPClient
#from pandac.PandaModules import Ramfile
#from pandac.PandaModules import URLSpec
#      
#http = HTTPClient()
#
#latitud=36.7
#longitud=-6.1
#channel = http.getDocument(URLSpec("http://www.openstreetmap.org/api/0.6/map?bbox=" + str(longitud-0.02)
#                                   +"," 
#                                   + str(latitud-0.02)
#                                   +","
#                                   + str(longitud+0.02)
#                                   +","
#                                   +str(latitud+0.02)))
#
#while channel.run():
#    print channel.getStatusString()
#
##if not channel.isDownloadComplete():
##    print 'error'
#
##channel.setConnectTimeout(15)
##channel.setHttpTimeout(15)
#rf = Ramfile()
#channel.downloadToRam(rf)
#
#if not channel.isValid():
#    print 'error2' 
# 
#data = rf.getData()
#print "got data:"
##print data

        
#M = Mapa('cadiz.osm')
#
#nodo_destino = '84807716'
#print nodo_destino
#print 'MatrizCamino1',M.getListaNodosSiguientes(nodo_destino),M.getListaNodosAnteriores(nodo_destino)
#print 'MatrizCamino',M.getListaAdyacentes(nodo_destino)
            
#    
#M = Mapa("cadiz.osm")
#
#print 'inicio'
#print M.Ruta("298900678","324013688")
#print 'fin'

#print 'total ways',M.getListaWays().__len__()
#print 'inicio'
#print M.getPorcionMapa(VBase3(0,0,0),Escalas['m-px']*500).keys().__len__()
#print 'fin'

#camino = M.getMatrizCamino()
#for p in camino.Mcamino:
#    if p == '259181304':
#        print 'nodo',p
#        print 'Siguientes: ',M.getListaNodosSiguientes(p)
#        print 'Anteriores: ',M.getListaNodosAnteriores(p)
#        print '------------------------------------------'
#    if p == '309847062':
#        print 'nodo',p
#        print 'Siguientes: ',M.getListaNodosSiguientes(p)
#        print 'Anteriores: ',M.getListaNodosAnteriores(p)
#        print '------------------------------------------'
#""" 
#
#"""
#lista = [1,2,3,4,5,6,7,8]
## todos menos el ultimo
#print lista[:-1]
## print todos menos el primero
#print lista[1:]
#"""
